
from homealone.core import *

# Dummy test interface
class TestInterface(Interface):
    def __init__(self, name, interface=None, event=None, **params):
        Interface.__init__(self, name, interface=interface, event=event)

    def read(self, addr):
        return 0

    def write(self, addr, value):
        return True




import time
import threading
from homealone.interfaces.gpioInterface import *
from homealone import *

# shade states
shadeUp = 0
shadeDown = 1
shadeRaising = 2
shadeLowering = 3

travelTime = [15, 15, 12, 12]

class ShadeInterface(Interface):
    def __init__(self, name, interface=None, gpioPins=None, event=None):
        Interface.__init__(self, name, interface=interface, event=event)
        self.gpioPins = gpioPins
        if not self.gpioPins:
            self.gpioPins = [18, 23, 24, 25, 22, 27, 17, 4]
        self.timers = [None, None, None, None]
        self.lock = threading.Lock()

    def addSensor(self, sensor):
        Interface.addSensor(self, sensor)
        self.states[sensor.addr] = 0    # initialize state to 0

    def read(self, addr):
        try:
            return self.states[addr]
        except:
            return 0

    def write(self, addr, value):
        self.newValue = value
        self.states[addr] = value + 2  # moving
        self.sensorAddrs[addr].notify()
        debug('debugShades', self.name, "state", addr, self.states[addr])
        # cancel the timer if it is running
        if self.timers[addr]:
            self.timers[addr].cancel()
        with self.lock:
            # set the direction
            debug('debugShades', self.name, "direction", addr, value)
            self.interface.write(self.gpioPins[addr*2], value)
            # start the motion
            debug('debugShades', self.name, "motion", addr, 1)
            self.interface.write(self.gpioPins[addr*2+1], 1)
        # clean up and set the final state when motion is finished
        def doneMoving():
            with self.lock:
                # stop the motion
                debug('debugShades', self.name, "motion", addr, 0)
                self.interface.write(self.gpioPins[addr*2+1], 0)
                # reset the direction
                debug('debugShades', self.name, "direction", addr, 0)
                self.interface.write(self.gpioPins[addr*2], 0)
                self.states[addr] = self.newValue # done moving
                self.sensorAddrs[addr].notify()
                debug('debugShades', self.name, "state", addr, self.states[addr])
        self.timers[addr] = threading.Timer(travelTime[addr], doneMoving)
        self.timers[addr].start()



import syslog
import os
import time
import threading
import traceback

from .env import *

# log a message to syslog or stdout
def log(*args):
    message = args[0]+" "   # first argument is the object doing the logging
    for arg in args[1:]:
        message += arg.__str__()+" "
    if sysLogging:
        syslog.syslog(message)
    else:
        print(time.strftime("%b %d %H:%M:%S", time.localtime())+" "+message)

# log the traceback for an exception
def logException(name, ex):
    tracebackLines = traceback.format_exception(None, ex, ex.__traceback__)
    log(name+":")
    for tracebackLine in tracebackLines:
        log(tracebackLine)

# thread object that logs a stack trace if there is an uncaught exception
class LogThread(threading.Thread):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.runTarget = self.run
        self.run = self.logThread

    def logThread(self):
        try:
            self.runTarget()
        except Exception as ex:
            logException("thread "+threading.currentThread().name, ex)

# convenience function to create and start a thread
def startThread(name, target, **kwargs):
    thread = LogThread(name=name, target=target, **kwargs)
    thread.start()

# log a debug message conditioned on a specified global variable
def debug(*args):
    try:
        if debugEnable:   # global debug flag enables debugging
            if globals()[args[0]]:  # only log if the specified debug variable is True
                log(*args[1:])
    except KeyError:
        pass

# log a stack trace conditioned on a specified global variable
def debugTraceback(debugType, debugName):
    try:
        if debugEnable:   # global debug flag enables debugging
            if globals()[debugType]:
                s = inspect.stack()
                for f in s:
                    log(debugName, f[1], f[2], f[3], f[4])
    except KeyError:
        pass

# wait for the network to be available
def waitForDns():
    wasWaiting = False
    while True:
        try:
            metricsHost = socket.gethostbyname(localController)
            if wasWaiting:
                log("DNS is up")
            return
        except:
            log("Waiting for DNS")
            wasWaiting = True
            time.sleep(1)

# create a human friendly label from a camel case name
def labelize(name):
    label = ""
    for char in name:
        if label:
            if label[-1].islower() and (char.isupper() or char.isnumeric()):
                label += " "
        label += char.lower()
    return label.capitalize()

# get the value of a variable from a file
def getValue(fileName, item):
    return json.load(open(fileName))[item]



# application template

from .core import *
from .schedule import *
from .rest.restServer import *

class HomeAlone(object):
    def __init__(self, name, globals, publish=True, port=restServicePort, metrics=False):
        self.name = name
        self.globals = globals
        self.publish = publish
        self.port = port
        self.metrics = metrics
        self.event = threading.Event()
        self.resources = Collection("resources")
        self.schedule = Schedule("schedule")
        config(self.globals)

    def run(self):
        if list(self.schedule.keys()) != []:
            self.schedule.start()
        if self.publish:
            self.restServer = RestServer(self.name, self.resources, event=self.event, port=self.port)
            self.restServer.start()

    def interface(self, interface):
        self.globals[interface.name] = interface

    def control(self, control, publish=True):
        self.globals[control.name] = control
        if publish:
            self.resources.addRes(control)

    def task(self, task, publish=True):
        self.schedule.addTask(task)
        self.globals[task.name] = self.schedule[task.name]
        if publish:
            self.resources.addRes(task)

    def style(self, style, resources=[]):
        if resources == []:
            resources = list(self.resources.values())
        for resource in listize(resources):
            resource.type = style

    def group(self, group, resources=[]):
        if resources == []:
            resources = list(self.resources.values())
        for resource in listize(resources:
            resource.group = group

    def label(self, label=None, resources=[]):
        if resources == []:
            resources = list(self.resources.values())
        for resource in resources:
            resource.label = labelize(resource.name)

# Read configuration files and set global variables
def config(globals):
    try:
        for configFileName in os.listdir(confDir):
            try:
                with open(confDir+configFileName) as configFile:
                    configLines = [configLine.rstrip('\n') for configLine in configFile]
                debugConfOpen = True
                for configLine in configLines:
                    if (len(configLine) > 0) and (configLine[0] != "#"):
                        try:
                            (key, value) = configLine.split("=")
                            globals[key.strip()] = eval(value)
                            if globals["debugEnable"] and globals["debugConf"]:   # can't use debugging module because of circular references
                                if debugConfOpen:   # log the file open retroactively if debugConf got set
                                    log("config open", configFileName)
                                    debugConfOpen = False
                                log("config read", configLine)
                        except Exception as ex:
                            log("config", "error evaluating", configLine, str(ex))
            except Exception as ex:
                log("config", "error reading", confDir+configFileName, str(ex))
    except:
        log("config", "no config directory", confDir)



# Extra sensors and controls derived from basic classes

from homealone.core import *

# A collection of sensors whose state is on if any one of them is on
class SensorGroup(Sensor):
    def __init__(self, name, sensorList, **kwargs):
        Sensor.__init__(self, name, **kwargs)
        self.sensorList = sensorList

    def getState(self, missing=None):
        if self.interface:
            # This is a cached resource
            return Sensor.getState(self)
        else:
            groupState = 0
            for sensor in self.sensorList:
                sensorState = sensor.getState(missing=0)
                debug("debugSensorGroup", self.name, "sensor:", sensor.name, "state:", sensorState)
                groupState = groupState or sensorState    # group is on if any one sensor is on
            debug("debugSensorGroup", self.name, "groupState:", groupState)
            return groupState

    # dictionary of pertinent attributes
    def dict(self, expand=False):
        attrs = Sensor.dict(self)
        attrs.update({"sensorList": [sensor.__str__() for sensor in self.sensorList]})
        return attrs

    def __repr__(self):
        return "\n".join([sensor.__str__() for sensor in self.sensorList])

# A set of Controls whose state can be changed together
class ControlGroup(SensorGroup, Control):
    def __init__(self, name, controlList, stateList=[], stateMode=False, type="controlGroup", **kwargs):
        SensorGroup.__init__(self, name, controlList, type=type, **kwargs)
        Control.__init__(self, name, type=type, **kwargs)
        self.stateMode = stateMode  # which state to return: False = SensorGroup, True = groupState
        self.groupState = 0
        if stateList == []:
            self.stateList = [[0,1]]*(len(self.sensorList))
        else:
            self.stateList = stateList

    def getState(self, missing=None):
        if self.interface:
            # This is a cached resource
            return Sensor.getState(self)
        else:
            if self.stateMode:
                return self.groupState
            else:
                return SensorGroup.getState(self)

    def setState(self, state, wait=False):
        if self.interface:
            # This is a cached resource
            return Control.setState(self, state)
        else:
            debug('debugState', self.name, "setState ", state)
            self.groupState = int(state)  # use Cycle - FIXME
            # Run it asynchronously in a separate thread.
            def setGroup():
                debug('debugThread', self.name, "started")
                for controlIdx in range(len(self.sensorList)):
                    control = self.sensorList[controlIdx]
                    debug("debugControlGroup", self.name, "control:", control.name, "state:", self.groupState)
                    control.setState(self.stateList[controlIdx][self.groupState])
                debug('debugThread', self.name, "finished")
            self.setGroupThread = LogThread(name="setGroupThread", target=setGroup)
            self.setGroupThread.start()
            self.notify(state)
            return True

    # dictionary of pertinent attributes
    def dict(self, expand=False):
        attrs = Control.dict(self)
        attrs.update({"controlList": [sensor.__str__() for sensor in self.sensorList]})
        return attrs

# A Control whose state depends on the states of a group of Sensors
class SensorGroupControl(SensorGroup, Control):
    def __init__(self, name, sensorList, control, **kwargs):
        Control.__init__(self, name, **kwargs)
        SensorGroup.__init__(self, name, sensorList, **kwargs)
        self.type = "sensorGroupControl"
        self.control = control

    def getState(self, missing=None):
        if self.interface:
            # This is a cached resource
            return Sensor.getState(self)
        else:
            return self.control.getState()

    def setState(self, state):
        # set the control on if any of the sensors is on
        # set the control off only if all the sensors are off
        controlState = state
        for sensor in self.sensorList:
            controlState = controlState or sensor.getState()
        if self.interface:
            # This is a cached resource
            return Control.setState(self, controlState)
        else:
            debug("debugSensorGroupControl", self.name, "control:", self.control.name, "state:", state, "controlState:", controlState)
            self.control.setState(controlState)

    # dictionary of pertinent attributes
    def dict(self, expand=False):
        attrs = Control.dict(self)
        attrs.update({"sensorList": [sensor.__str__() for sensor in self.sensorList],
                      "control": self.control.__str__()})
        return attrs

# Calculate a function of a list of sensor states
class CalcSensor(Sensor):
    def __init__(self, name, sensors=[], function="", **kwargs):
        Sensor.__init__(self, name, **kwargs)
        type = "sensor"
        self.sensors = sensors
        self.function = function.lower()
        self.className = "Sensor"

    def getState(self, missing=0):
        value = 0
        try:
            if self.function in ["sum", "avg", "+"]:
                for sensor in self.sensors:
                    value += sensor.getState(missing=0)
                if self.function == "avg":
                    value /+ len(self.sensors)
            elif self.function in ["*"]:
                for sensor in self.sensors:
                    value *= sensor.getState(missing=0)
            elif self.function in ["diff", "-"]:
                value = self.sensors[0].getState(missing=0) - self.sensors[1].getState(missing=0)
        except Exception as ex:
            logException(self.name, ex)
        return value

# Sensor that only reports its state if all the specified resources are in the specified states
class DependentSensor(Sensor):
    def __init__(self, name, interface, sensor, conditions, default=off, **kwargs):
        Sensor.__init__(self, name, **kwargs)
        type = "sensor"
        self.className = "Sensor"
        self.sensor = sensor
        self.conditions = conditions
        self.default = default

    def getState(self, missing=0.0):
        debug('debugState', self.name, "getState")
        for (sensor, condition, value) in self.conditions:
            sensorState = sensor.getState()
            sensorName = sensor.name
            if isinstance(value, Sensor):
                value = value.getState()
            debug('debugDependentControl', self.name, sensorName, sensorState, condition, value)
            try:
                if eval(str(sensorState)+condition+str(value)):
                    return self.sensor.getState()
                else:
                    return missing
            except Exception as ex:
                log(self.name, "exception evaluating condition", str(ex))
                return missing

# Control that can only be turned on if all the specified resources are in the specified states
class DependentControl(Control):
    def __init__(self, name, interface, control, conditions, **kwargs):
        Control.__init__(self, name, **kwargs)
        type = "control"
        self.className = "Control"
        self.control = control
        self.conditions = conditions

    def getState(self, missing=None):
        return self.control.getState()

    def setState(self, state, wait=False):
        debug('debugState', self.name, "setState ", state)
        for (sensor, condition, value) in self.conditions:
            sensorState = sensor.getState()
            sensorName = sensor.name
            if isinstance(value, Sensor):
                value = value.getState()
            debug('debugDependentControl', self.name, sensorName, sensorState, condition, value)
            try:
                if eval(str(sensorState)+condition+str(value)):
                    self.control.setState(state)
            except Exception as ex:
                log(self.name, "exception evaluating condition", str(ex))

# Control that can be set on but reverts to off after a specified time
class MomentaryControl(Control):
    def __init__(self, name, interface, addr=None, duration=1, group="", type="control", location=None, label="", event=None):
        Control.__init__(self, name, interface, addr, group=group, type=type, location=location, label=label, event=event)
        self.className = "Control"
        self.duration = duration
        self.timedState = 0
        self.timer = None

    def setState(self, state, wait=False):
        # timeout is the length of time the control will stay on
        debug("debugState", "OneShotControl", self.name, "setState", state)
        if not self.timedState:
            self.timedState = state
            if self.interface:
                self.interface.write(self.addr, self.timedState)
            self.timer = threading.Timer(self.duration, self.timeout)
            self.timer.start()
            debug("debugState", "OneShotControl", self.name, "timer", self.timedState)
            self.notify()

    def timeout(self):
        self.timedState = 0
        debug("debugState", "OneShotControl", self.name, "timeout", self.duration)
        debug("debugState", "OneShotControl", self.name, "setState", self.timedState)
        if self.interface:
            self.interface.write(self.addr, self.timedState)
        self.notify()

    def getState(self, missing=None):
        return self.timedState

# Control that has a specified list of values it can be set to
class MultiControl(Control):
    def __init__(self, name, interface, addr=None, values=[], **kwargs):
        Control.__init__(self, name, interface, addr, **kwargs)
        type = "control"
        self.className = "MultiControl"
        self.values = values

    def setState(self, state, wait=False):
        debug("debugState", "MultiControl", self.name, "setState", state, self.values)
        if state in self.values:
            return Control.setState(self, state)
        else:
            return False

    # dictionary of pertinent attributes
    def dict(self, expand=False):
        attrs = Control.dict(self)
        attrs.update({"values": self.values})
        return attrs

# Control that has specified numeric limits on the values it can be set to
class MinMaxControl(Control):
    def __init__(self, name, interface, addr=None, minValue=0, maxValue=1, **kwargs):
        Control.__init__(self, name, interface, addr, **kwargs)
        type = "control"
        self.className = "Control"
        self.setMinMax(minValue, maxValue)

    def setState(self, state, wait=False):
        state = int(state)
        debug("debugState", "MinMaxControl", self.name, "setState", state, self.minValue, self.maxValue)
        if state < self.minValue:
            value = self.minValue
        elif state > self.maxValue:
            value = self.maxValue
        else:
            value = state
        Control.setState(self, value)

    def setMinMax(self, minValue, maxValue):
        self.minValue = minValue
        self.maxValue = maxValue

# Sensor that captures the minimum state value of the specified sensor
class MinSensor(Sensor):
    def __init__(self, name, interface, addr, sensor, **kwargs):
        Sensor.__init__(self, name, interface, addr, **kwargs)
        type = "sensor"
        self.className = "Sensor"
        self.sensor = sensor
        try:
            self.minState = self.interface.read(self.addr)
        except:
            self.minState = 999

    def getState(self, missing=None):
        if self.interface:
            if self.interface.__class__.__name__ == "RestInterface":
                return self.interface.read(self.addr)
            else:
                self.minState = self.interface.read(self.addr)
        sensorState = self.sensor.getState()
        if sensorState < self.minState:
            if sensorState != 0:    # FIXME
                self.minState = sensorState
                if self.interface:
                    self.interface.write(self.addr, self.minState)
        return self.minState

    # reset the min value
    def setState(self, value):
        self.minState = value
        if self.interface:
            self.interface.write(self.addr, self.minState)

    # dictionary of pertinent attributes
    # def dict(self, expand=False):
    #     attrs = Control.dict(self)
    #     attrs.update({"sensor": str(self.sensor)})
    #     return attrs

# Sensor that captures the maximum state value of the specified sensor
class MaxSensor(Sensor):
    def __init__(self, name, interface, addr, sensor, **kwargs):
        Sensor.__init__(self, name, interface, addr, **kwargs)
        type = "sensor"
        self.className = "Sensor"
        self.sensor = sensor
        try:
            self.maxState = self.interface.read(self.addr)
        except:
            self.maxState = 0

    def getState(self, missing=0):
        if self.interface:
            if self.interface.__class__.__name__ == "RestInterface":
                return self.interface.read(self.addr)
            else:
                self.maxState = self.interface.read(self.addr)
        sensorState = self.sensor.getState()
        if sensorState > self.maxState:
            self.maxState = sensorState
            if self.interface:
                self.interface.write(self.addr, self.maxState)
        return self.maxState

    # reset the max value
    def setState(self, value):
        self.maxState = value
        if self.interface:
            self.interface.write(self.addr, self.maxState)

    # dictionary of pertinent attributes
    # def dict(self, expand=False):
    #     attrs = Control.dict(self)
    #     attrs.update({"sensor": str(self.sensor)})
    #     return attrs

# Sensor that captures the accumulated state values of the specified sensor
class AccumSensor(Sensor):
    def __init__(self, name, interface, sensor, multiplier=1, **kwargs):
        Sensor.__init__(self, name, interface, addr, **kwargs)
        type = "sensor"
        self.className = "Sensor"
        self.sensor = sensor
        self.multiplier = multiplier
        try:
            self.accumValue = self.interface.read(self.name)
        except:
            self.accumValue = 0

    def getState(selfmissing=0):
        self.accumValue = self.sensor.getState() * self.multiplier
        if self.interface:
            self.interface.write(self.name, self.accumValue)
        return self.accumValue

    # reset the accumulated value
    def setState(self, value):
        self.accumValue = value
        if self.interface:
            self.interface.write(self.name, self.accumValue)

# sensor that returns the value of an attribute of a specified sensor
class AttributeSensor(Sensor):
    def __init__(self, name, interface, addr, sensor, attr, **kwargs):
        Sensor.__init__(self, name, interface, addr, **kwargs)
        type = "sensor"
        self.sensor = sensor
        self.attr = attr

    def getState(self, missing=None):
        return getattr(self.sensor, self.attr)

    # dictionary of pertinent attributes
    def dict(self, expand=False):
        attrs = Sensor.dict(self)
        attrs.update({"sensor": str(self.sensor),
                      "attr": self.attr})
        return attrs

# a sensor that is proxied from another server
class ProxySensor(Sensor):
    def __init__(self, name, resources, **kwargs):
        Sensor.__init__(self, name, **kwargs)
        self.resources = resources

    def getState(self, missing=None):
        try:
            state = self.resources[self.name].getState(missing=missing)
            self.enable()
            return state
        except KeyError:
            self.disable()
            return missing

# a control that is proxied from another server
class ProxyControl(ProxySensor):
    def __init__(self, name, resources, **kwargs):
        ProxySensor.__init__(self, name, resources, **kwargs)
        self.resources = resources

    def setState(self, value, **kwargs):
        try:
            return self.resources[self.name].setState(value, **kwargs)
        except KeyError:
            return False



tempWatchInterval = 1

# states
enabled = 1

# unit types
unitTypeHeater = 0
unitTypeAc = 1

from homealone import *

# a temperature controlled heating or cooling unit
class TempControl(Control):
    def __init__(self, name, interface, unitControl, tempSensor,
                tempTargetControl=None, unitType=0, hysteresis=[1, 1],
                **kwargs):
        Control.__init__(self, name, interface, **kwargs)
        self.className = "Control"
        self.unitControl = unitControl              # the control for the heating or cooling unit
        self.tempSensor = tempSensor                # associated temp sensor
        self.tempTargetControl = tempTargetControl  # the control that sets the temp target
        self.tempTarget = 0                         # current temp target value
        self.unitType = unitType                    # type of unit: 0=heater, 1=cooler
        self.hysteresis = hysteresis                # how much to overshoot or undershoot the temp target [-low, +high]
        self.controlState = off                     # state of this control
        self.inhibited = False                      # this control is inhibited

    def getState(self, wait=False, missing=None):
        debug('debugState', self.name, "getState ", self.controlState)
        return self.controlState

    def setState(self, state, wait=False):
        debug('debugState', self.name, "setState ", state)
        # thread to monitor the temperature
        def tempWatch():
            debug('debugTempControl', self.name, "tempWatch started")
            while self.controlState == enabled:  # stop when state is set to off
                time.sleep(tempWatchInterval)
                currentTemp = int(self.tempSensor.getState() + .5)
                if currentTemp > 0:                 # don't do anything if no temp reading
                    if self.tempTargetControl:
                        try:
                            self.tempTarget = int(self.tempTargetControl.getState())
                        except ValueError:
                            self.tempTarget = 0
                    debug('debugTempControlTemp', self.name, "currentTemp:", currentTemp, "targetTemp:", self.tempTarget, "inhibit:", self.inhibited)
                    if self.inhibited or \
                        ((self.unitType == unitTypeHeater) and (currentTemp >= self.tempTarget + self.hysteresis[1])) or \
                        ((self.unitType == unitTypeAc) and (currentTemp <= self.tempTarget - self.hysteresis[0])):
                        # unit is inhibited or the target temp has been reached
                        if self.unitControl.getState() != off:
                            # turn the unit off
                            self.unitControl.setState(off)
                            debug('debugTempControl', self.name, "unit off")
                    elif not self.inhibited and \
                        (((self.unitType == unitTypeHeater) and (currentTemp <= self.tempTarget - self.hysteresis[0])) or \
                        ((self.unitType == unitTypeAc) and (currentTemp >= self.tempTarget + self.hysteresis[1]))):
                        # unit is not inhibited and the temp is sufficiently off target
                        if self.unitControl.getState() != on:
                            # turn the unit on
                            self.unitControl.setState(on)
                            debug('debugTempControl', self.name, "unit on")
                    else:
                        # do nothing
                        pass
            if self.unitControl.getState() != off:
                # turn the unit off
                self.unitControl.setState(off)
                debug('debugTempControl', self.name, "unit off")
            debug('debugTempControl', self.name, "tempWatch terminated")
        self.controlState = state
        if self.controlState == enabled:      # start the monitor thread when state set to enabled
            tempWatchThread = LogThread(name="tempWatchThread", target=tempWatch)
            tempWatchThread.start()
        self.notify()

    def setTarget(self, tempTarget, hysteresis=[1, 1], wait=False):
        debug('debugTempControl', self.name, "setTarget ", tempTarget, hysteresis)
        self.tempTarget = tempTarget
        self.hysteresis = hysteresis

    def setInhibit(self, value):
        debug('debugTempControl', self.name, "inhibit", value)
        self.inhibited = value



import os
import socket

# directory structure
hostname = socket.gethostname()
rootDir = os.path.expanduser("~")+"/"
confDir = rootDir+"conf/"
keyDir = rootDir+"keys/"
stateDir = rootDir+"state/"
soundDir = rootDir+"sounds/"
dataLogDir = rootDir+"data/"
dataLogFileName = ""

# global variables that must be set here
sysLogging = False
debugEnable = False

# Localization - define these in the config file
latLong = (0.0, 0.0)
elevation = 0 # elevation in feet
tempScale = "F"
localController = "localhost"
